searchData={"items":[{"type":"module","title":"ExSleeplock","doc":"API allowing creation of locks to throttle concurrent processing via\nobtaining a lock before executing functionality you want to limit.\n\nBefore a lock can be used it must be created using `new/2`. When `new/2`\nis called a process is created to manage the lock. The lock is identified\nby a unique atom and the number of slots indicates how many processes can\nhold the lock at once. The name of the lock is the process name.\n\nThe preferred way to use a created lock is with the function `execute/2`.\nThis function acquires the lock, executes the function passed in and\nthen releases the lock (returning the function return value to the caller)\nThis ensures that the lock is always released even if an exception is raised.","ref":"ExSleeplock.html"},{"type":"function","title":"ExSleeplock.acquire/1","doc":"Aquire a lock\n\nThis will block until a lock can be acquired. When a lock is\nacquired the caller is responsible for calling `release/1`.","ref":"ExSleeplock.html#acquire/1"},{"type":"function","title":"Arguments - ExSleeplock.acquire/1","doc":"* name - a unique atom identifying the lock","ref":"ExSleeplock.html#acquire/1-arguments"},{"type":"function","title":"Returns - ExSleeplock.acquire/1","doc":"* `:ok` - after lock is acquired\n* `{:error, :lock_not_found}` - if called with a name that doesn't match\n  an existing lock","ref":"ExSleeplock.html#acquire/1-returns"},{"type":"function","title":"ExSleeplock.attempt/1","doc":"A non-blocking version of `acquire/1`","ref":"ExSleeplock.html#attempt/1"},{"type":"function","title":"Arguments - ExSleeplock.attempt/1","doc":"* name - a unique atom identifying the lock","ref":"ExSleeplock.html#attempt/1-arguments"},{"type":"function","title":"Returns - ExSleeplock.attempt/1","doc":"* `:ok` - lock is acquired\n* `{:error, :unavailable}` - lock is not acquired\n* `{:error, :lock_not_found}` - if called with a name that doesn't match\n  an existing lock","ref":"ExSleeplock.html#attempt/1-returns"},{"type":"function","title":"ExSleeplock.execute/2","doc":"Execute the function passed in by first acquiring the lock\n\nThe lock is automatically released when the function completes. This call\nblocks until a lock is acquired.","ref":"ExSleeplock.html#execute/2"},{"type":"function","title":"Arguments - ExSleeplock.execute/2","doc":"* name - a unique atom identifying the lock\n* fun - a function that executes after the lock is acquired","ref":"ExSleeplock.html#execute/2-arguments"},{"type":"function","title":"Returns - ExSleeplock.execute/2","doc":"* on success the call returns the value returned by the function supplied\n* `{:error, :lock_not_found}` - if called with a name that doesn't match\n  an existing lock","ref":"ExSleeplock.html#execute/2-returns"},{"type":"function","title":"ExSleeplock.help/0","doc":"Return help message","ref":"ExSleeplock.html#help/0"},{"type":"function","title":"ExSleeplock.lock_state/1","doc":"Return lock's current state","ref":"ExSleeplock.html#lock_state/1"},{"type":"function","title":"ExSleeplock.new/2","doc":"Create a lock","ref":"ExSleeplock.html#new/2"},{"type":"function","title":"Arguments - ExSleeplock.new/2","doc":"* name - a unique atom identifying the lock. The name becomes the process name\n  for the lock\n* num_slots - a positive integer indicating how many processes are allowed to hold\n  this lock at once","ref":"ExSleeplock.html#new/2-arguments"},{"type":"function","title":"Returns - ExSleeplock.new/2","doc":"* `{:ok, pid}` - on success a GenServer using the name supplied is started. This\n   process is supervised by the library.\n* `{:error, :invalid, msg}` - returned when parameters are invalid\n* `{:error, {:already_started, pid}}` - returned when attempting to create the same\n  lock more than once","ref":"ExSleeplock.html#new/2-returns"},{"type":"function","title":"ExSleeplock.release/1","doc":"Release an acquired lock","ref":"ExSleeplock.html#release/1"},{"type":"function","title":"Arguments - ExSleeplock.release/1","doc":"* name - a unique atom identifying the lock","ref":"ExSleeplock.html#release/1-arguments"},{"type":"function","title":"Returns - ExSleeplock.release/1","doc":"* `:ok` - always returned even if you do not have an acquired lock\n* `{:error, :lock_not_found}` - if called with a name that doesn't match\n  an existing lock","ref":"ExSleeplock.html#release/1-returns"},{"type":"type","title":"ExSleeplock.lock_info/0","doc":"General information about a lock","ref":"ExSleeplock.html#t:lock_info/0"},{"type":"type","title":"ExSleeplock.lock_state/0","doc":"Number of processes that have obtained the lock and are currently running and\nthe number of processes waiting for a lock","ref":"ExSleeplock.html#t:lock_state/0"},{"type":"extras","title":"ExSleeplock","doc":"# ExSleeplock\n\nAllow concurrent throttling using a named lock in Elixir.\n\nThis project was inspired by the Erlang project\n[sleeplocks](https://hex.pm/packages/sleeplocks) (thanks to _Isaac Whitfield_\nwho created the Erlang sleeplocks library). It provides similar functionality\nbut adds:\n\n- monitoring of processes that take locks\n- dynamic supervision of the locks themselves\n- telemetry\n- creation of locks at startup from config file\n- function to get the current state of a lock (number processes running and waiting)","ref":"readme.html"},{"type":"extras","title":"Overview - ExSleeplock","doc":"This library provides an app with the ability to create a named concurrent lock.\nThe app calls a library function to create a lock with a unique (atom) name and\na level of concurrency (positive integer). When a lock is created the library\nstarts a process using the atom as it's name. The level of concurrency is\ndefined by the \"num_slots\". So, `ExSleeplock.new(:my_lock, _num_slots = 2)`\ncreates a concurrent lock named `:my_lock` that allows two processes to have a\nlock concurrently.\n\nThe app executes a block of code protected by the lock. The lock is kept until the\nblock of code completes. Then it is released. The easiest (and suggested) way for\nan app to do this is:\n\n```\nExSleeplock.execute(:my_lock, fn -> some_function() end)\n```\n\nWhen the maximum number of concurrent processes are running there are no \"slots\"\navailable. If a caller attempts to obtain the lock to execute code the caller is\nplaced in a queue. The execution of that process is suspended until a lock is\navailable. Then the process locks and continues its execution. This requires no\nintervention by the app. Waiting processes are in a queue. They are handled\nin a FIFO (first in first out) manner.\n\nThe API available is defined in `ExSleeplock`:\n\n* `ExSleeplock.new/2` - create a lock. Before a lock can be used it\n  has to be created. This is generally done when the application starts.\n* `ExSleeplock.execute/2` - execute a function after acquiring a lock and\n  release it when function completes.\n* `ExSleeplock.acquire/1` - acquire a lock. Blocks until lock is acquired. Caller\n  is responsible for releasing the lock.\n* `ExSleeplock.attempt/1` - attempt to acquire a lock. Doesn't block. Returns\n  `{:error, :unavailable}` if lock cannot be obtained.\n* `ExSleeplock.release/1` - release a lock","ref":"readme.html#overview"},{"type":"extras","title":"Using Locks - ExSleeplock","doc":"```\nExSleeplock.new(:process_foo, _num_slots = 2)\n```\n\nThis creates a lock called `:process_foo` with two slots. Only two processes\nare allowed to execute concurrently. Using it from an app looks something\nlike this:\n\n```\nresult = ExSleeplock.execute(:process_foo, fn -> some_work() end)\n```\n\nIf there are already two processes with a `:process_foo` lock the third\nprocess is placed in a queue and resumes execution when one of the two currently\nrunning processes unlock.\n\nUsing `ExSleeplock.execute/2` is the easiest way to use this library but you can\nmanage your own lock and timing of execution of a block of code using\n`ExSleeplock.acquire/1` (blocking) or `ExSleeplock.attempt/1` (non-blocking). If you\nare using the calls in this way then you are responsible for ensuring\n`ExSleeplock.release/1` is called. Until `ExSleeplock.release/1` is called a slot is\nbeing used and a new process may not be able to move forward.\n\nWhen a lock is obtained by a process the library creates a monitor on the process\ntaking the lock. If the process exits unexpectedly the lock is automatically\nreleased (even though the processs never called `ExSleeplock.release/1`).\n\nYou can only obtain a lock once for a process (pid). Once a process has a lock\ncalling a function to obtain a lock again always succeeds. However, no telemetry\nevent is generated.","ref":"readme.html#using-locks"},{"type":"extras","title":"Configuring Lock Creation on Startup - ExSleeplock","doc":"When the library starts up it reads the application environment to see if any\nlocks are configured. These locks are automatically started. For example:\n\n```\nconfig :ex_sleeplock, locks: [%{name: :test1, num_slots: 2}, %{name: :test2, num_slots: 3}]\n```\n\nThis config would create two locks named `:test1` and `:test2` with the indicated\nconcurrency.\n\nUsing this approach is not required. An application can create a lock by calling\nthe function `ExSleeplock.new/2`. This should be done when the application starts.","ref":"readme.html#configuring-lock-creation-on-startup"},{"type":"extras","title":"Telemetry - ExSleeplock","doc":"The library does not generate any telemetry events by default. The app can\nconfigure it to do so by setting up the following environment (config file)\nelement:\n\n```\n:ex_sleeplock, notifier: ExSleeplock.EventGenerator.LockTelemetry\n```\n\nWhen this is setup the following telemetry events are generated.\n\n- `[:ex_sleeplock, :lock_created]` - lock created\n  - measurements - `%{value: 1}`\n  - metadata - `t:ExSleeplock.lock_info/0`\n- `[:ex_sleeplock, :lock_acquired]` - lock acquired\n  - measurements - `t:ExSleeplock.lock_state/0`\n  - metadata - `t:ExSleeplock.lock_info/0`\n- `[:ex_sleeplock, :lock_released]` - lock released\n  - measurements - `t:ExSleeplock.lock_state/0`\n  - metadata - `t:ExSleeplock.lock_info/0`\n- `[:ex_sleeplock, :lock_waiting]` - a process is added to queue to wait\n  - measurements - `t:ExSleeplock.lock_state/0`\n  - metadata - `t:ExSleeplock.lock_info/0`\n\nSee the documentation for `ExSleeplock` for more information.","ref":"readme.html#telemetry"},{"type":"extras","title":"Sample Use Case - ExSleeplock","doc":"The library is useful when you can have code within your app that must access a\nresource that is limited in some way. You want only 2 or 3 (or some limited\nnumber) of processes accessing the resource at the same time.\n\nAn example use case is an app that provides an API and processes messages from\nKafka. In this case:\n\n- The API performs database reads and writes\n- The Kafka messages are processed in parallel by different processes and the\n  processing performs database reads and writes\n\nThere are two issues in this use case. First, there are a limited number of\nconnections to the database. The app can get more incoming messages in parallel\nthen the number of available database connecitons. Second, the app cannot use\nall the connections to process incoming messages. If a connection is not\navailable the API cannot respond in a timely manner to incoming API requests.","ref":"readme.html#sample-use-case"},{"type":"extras","title":"Development Session - ExSleeplock","doc":"Start an iex session and paste the following module into the session.\nThe `process/3` function in the module simulates processing by sleeping\nfor caller specified number of milliseconds plus some random number\nof milliseconds (< 100) to provide a bit of randomness.\n\n```\ndefmodule TestSleepwalk do\n  def process(name, wait_time_in_ms, idx) do\n    IO.puts(\"Queued  [#{idx}] at: #{current_time()}.\")\n    ExSleeplock.execute(name, fn ->\n      # Simulate procesesing when a lock is obtained\n      IO.puts(\"Lock    [#{idx}] at: #{current_time()}.\")\n      Process.sleep(wait_time_in_ms + :rand.uniform(100))\n      IO.puts(\"Release [#{idx}] at: #{current_time()}.\")\n      idx\n    end)\n  end\n\n  defp current_time do\n    DateTime.utc_now() |> DateTime.truncate(:millisecond)\n  end\nend\n```\n\nNow create a lock called `:test_sleepwalk` that allows 2 concurrent processes.\n\n```\niex> {:ok, pid} = ExSleeplock.new(:test_sleepwalk, 2)\n```\n\nNow let's start more than 2 processes using Task.async. All of the processes call\n`TestSleepwalk.process/3`. An explanation of what happens is provided below.\n\n```\niex> results = Enum.map(1..6, fn idx -> Task.async(fn -> TestSleepwalk.process(:test_sleepwalk, 1_000, idx) end) end) |> Task.await_many(:infinity)\n\n# All the tasks get queued. Before a call is done to obtain a lock the function outputs\n# an index number. This means that all of this output happens first.\n\nQueued  [1] at: 2024-06-12 16:22:43.072Z.\nQueued  [2] at: 2024-06-12 16:22:43.072Z.\nQueued  [3] at: 2024-06-12 16:22:43.072Z.\nQueued  [4] at: 2024-06-12 16:22:43.072Z.\nQueued  [5] at: 2024-06-12 16:22:43.072Z.\nQueued  [6] at: 2024-06-12 16:22:43.072Z.\n\n# Now the tasks attempt to obtain a lock and execute. Only two are\n# allowed to obtain a lock. The rest are waiting.\n\nLock    [2] at: 2024-06-12 16:22:43.072Z.\nLock    [1] at: 2024-06-12 16:22:43.072Z.\n\n# One of the two processes that started complete and release their lock\n# This allows another process to get the lock and begin executing.\n\nRelease [1] at: 2024-06-12 16:22:44.102Z.\nLock    [3] at: 2024-06-12 16:22:44.103Z.\n\n# Second task completes and the next waiting task locks and begins\n# executing\n\nRelease [2] at: 2024-06-12 16:22:44.102Z.\nLock    [4] at: 2024-06-12 16:22:44.103Z.\n\n# You should be able to see what is happening now. The remainder of\n# the output is:\n\nRelease [3] at: 2024-06-12 16:22:45.152Z.\nLock    [5] at: 2024-06-12 16:22:45.153Z.\nRelease [4] at: 2024-06-12 16:22:45.188Z.\nLock    [6] at: 2024-06-12 16:22:45.190Z.\nRelease [5] at: 2024-06-12 16:22:46.202Z.\nRelease [6] at: 2024-06-12 16:22:46.246Z.\n[1, 2, 3, 4, 5, 6]\n```\n\nYou can use `ExSleeplock.lock_state/1` to examine the current lock state.\n\n```\niex> ExSleepLock.lock_state(:test_sleepwalk)\n%{running: 0, waiting: 0}\n```","ref":"readme.html#development-session"},{"type":"extras","title":"Contributing - ExSleeplock","doc":"Pull requests are welcome. The project uses Github Actions. All PR's\nshould:\n\n- have no test failures\n- be formatted properly (`mix format`)\n- have no credo issues (`mix credo --strict`)\n- have no dialyzer issues (`mix dialyzer`)\n\nIf new functionality is proposed there should be unit tests to cover it.","ref":"readme.html#contributing"},{"type":"extras","title":"License - ExSleeplock","doc":"MIT","ref":"readme.html#license"},{"type":"extras","title":"Changelog","doc":"0.10.2\n------\n#### Fixes\n* Push tag before hex publish so code aligns with doc in hex.pm\n\n\n0.10.0\n------\n#### Enhancements\n* Better test coverage\n* Update README.md\n* Set mix version for hex.pm","ref":"changelog.html"},{"type":"extras","title":"License","doc":"MIT License\n\nCopyright (c) 2024 Frank McGeough\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the \"Software\"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE\nSOFTWARE.","ref":"license.html"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}